/*	XGFileSpecifier.cpp
 *
 *		This is the file specifier object; this encapsulates the file
 *	and directory information for a file in a universally accessable
 *	object
 */

/**********************************************************************

  YAAF is Copyright 1997-1999 by William Woody.

  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions 
  are met:

  Redistributions of source code must retain the above copyright notice, 
  this list of conditions and the following disclaimer.

  Neither the name of the copyright holder or other contributors may
  be used to promote products derived from this software without
  specific written prior permission.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
  OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS 
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN 
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
  SOFTWARE.
  
  Contact:   woody@alumni.caltech.edu
  Web page:  http://www.pandawave.com/yaaf/

**********************************************************************/

#include <XConfig.h>
#include <XStringUtil.h>
#include <XError.h>
#include <XFile.h>
#include <XWindow.h>
#include <string.h>
#include <ctype.h>
#include <XApplication.h>


#if TARGET_CARBON
static NavEventUPP GEventProcPtr;
#endif


/************************************************************************/
/*																		*/
/*	Construction/Destruction											*/
/*																		*/
/************************************************************************/

#if OPT_MACOS

/*$	XGFileSpecifier::XGFileSpecifier
 *
 *		Construct this file specifier from the data provided.
 */

XGFileSpecifier::XGFileSpecifier(FSSpec &f)
{
	fFileName = f;
}

#else

/*$	XGFileSpecifier::XGFileSpecifier
 *
 *		Construct this file specifier from the data provided.
 */

XGFileSpecifier::XGFileSpecifier(const char *f)
{
	fFileName = f;
}

#endif

/*$	XGFileSpecifier::XGFileSpecifier
 *
 *		Create an 'empty' file specifier. You must use a copy
 *	operation to initialize this before using.
 */

XGFileSpecifier::XGFileSpecifier()
{
#if OPT_MACOS
	fFileName.name[0] = 0;
#endif
}

/*$	XGFileSpecifier::XGFileSpecifier
 *
 *		Construct this file specifier from an alias handle.
 */

XGFileSpecifier::XGFileSpecifier(XGHandle &alias)
{
	(*this) = alias;
}

XGFileSpecifier::XGFileSpecifier(const XGFileSpecifier& alias)
{
	this->fFileName = alias.fFileName;
}

/*$	XGFileSpecifier::~XGFileSpecifier
 *
 *		Deconstruct me
 */

XGFileSpecifier::~XGFileSpecifier()
{
}

/************************************************************************/
/*																		*/
/*	Copy operators														*/
/*																		*/
/************************************************************************/

/*$	XGFileSpecifier::operator = 
 *
 *		Copy an OS specific file specifier into this object.
 */

#if OPT_MACOS
XGFileSpecifier &XGFileSpecifier::operator = (const FSSpec &f)
{
	fFileName = f;
	return *this;
}
#else
XGFileSpecifier &XGFileSpecifier::operator = (const char *f)
{
	fFileName = f;
	return *this;
}
#endif

/*$	XGFileSpecifier::operator =
 *
 *		Create this specifier from the alias record provided.
 */

XGFileSpecifier &XGFileSpecifier::operator = (XGHandle &alias)
{
#if OPT_MACOS
	extern long _GMacOSVersion;
	
	if ((_GMacOSVersion < 0x0700) || (alias.GetSize() == sizeof(FSSpec))) {
		/*
		 *	Either we only have a file specifier or we're running
		 *	an earlier version than System 7
		 */
		
		alias.GetMemory(0,sizeof(fFileName),(void *)&fFileName);
	} else {
		/*
		 *	We've got an alias...
		 */
		
		Handle h;
		long l;
		OSErr err;
		Boolean change;
		
		l = alias.GetSize() - sizeof(FSSpec);
		h = NewHandle(l);
		Assert(h != NULL);
		HLock(h);
		alias.GetMemory(sizeof(FSSpec),l,*h);
		HUnlock(h);
		
		err = ResolveAlias(NULL,(AliasHandle)h,&fFileName,&change);
		DisposeHandle(h);
		Assert(err == noErr);
	}
#else
	XGHandleLock lock(alias);
	
	fFileName = (char *)*alias;
#endif

	return *this;
}

/*$	XGFileSpecifier::operator =
 *
 *		Copy from file specifier
 */

XGFileSpecifier &XGFileSpecifier::operator = (const XGFileSpecifier &data)
{
	fFileName = data.fFileName;
	return *this;
}

bool XGFileSpecifier::operator == (const XGFileSpecifier &data) const
{
#if OPT_MACOS
	return data.fFileName.vRefNum == fFileName.vRefNum &&
			data.fFileName.parID == fFileName.parID &&
			!strcmp( (const char *)data.fFileName.name, (const char *)fFileName.name );
#else
	return data.fFileName == fFileName;
#endif
}


/************************************************************************/
/*																		*/
/*	Access operators													*/
/*																		*/
/************************************************************************/

/*$	XGFileSpecifier::GetAlias
 *
 *		Get the alias for this file. This alias provides a mechanism for
 *	finding this file later. This alias is only useful on one machine,
 *	but provids a way of finding that file later.
 */

void XGFileSpecifier::GetAlias(XGHandle &alias) const
{
#if OPT_MACOS
	extern long _GMacOSVersion;
	
	if (_GMacOSVersion < 0x0700) {
		alias.SetSize(sizeof(FSSpec));
		alias.SetMemory(0,sizeof(FSSpec),(char *)&fFileName);
	} else {
		AliasHandle h;
		OSErr err;
		long l;
		
		err = NewAlias(NULL,&fFileName,&h);

		Assert(err == noErr);
		l = GetHandleSize((Handle)h);
		alias.SetSize(sizeof(FSSpec) + l);

		HLock((Handle)h);
		alias.SetMemory(0,sizeof(FSSpec),(void *)&fFileName);
		alias.SetMemory(sizeof(FSSpec),l,(void *)*h);
		DisposeHandle((Handle)h);
	}
#else
	short len;
	const char *str;
	
	str = fFileName.Get();
	len = strlen(str) + 1;
	alias.SetSize(len);
	strcpy((char *)*alias,str);
#endif
}

/*$	XGFileSpecifier::OpenFile
 *
 *		Open this file using the ANSI access routines. This provides
 *	a way to use the standard ANSI file access routines fprintf(),
 *	fread(), fwrite(), etc., when reading/writing this file. Normally
 *	you wouldn't use this routine, but there it is anyways.
 */

FILE *XGFileSpecifier::OpenFile(const char *access) const
{
#if OPT_WINOS || OPT_XWIN
	return fopen(fFileName.Get(),access);
#else

#if TARGET_CARBON
	/*
	 *	Carbonized MacOS version sets the default volume first
	 */
	
	FILE *f;
	char name[64];
	short vRefNum;
	long dirID;
	
	HGetVol(NULL,&vRefNum,&dirID);
	HSetVol(NULL,fFileName.vRefNum,fFileName.parID);
	
	GetName(name);
	f = fopen(name,access);
	
	HSetVol(NULL,vRefNum,dirID);
	
	return f;
#else
	/*
	 *	System 6/7 friendly MacOS version sets the default volume first
	 */
	
	FILE *f;
	char name[64];
	short wdRefNum;
	OSErr err;
	short vRefNum;
	
	err = OpenWD(fFileName.vRefNum,fFileName.parID,'ERIK',&wdRefNum);
	if (err) return NULL;
	GetVol(NULL,&vRefNum);
	SetVol(NULL,wdRefNum);
	
	GetName(name);
	f = fopen(name,access);
	
	SetVol(NULL,vRefNum);
	CloseWD(wdRefNum);
	
	return f;
#endif

#endif
}
/*$	XGFileSpecifier::SetFile
 *
 *		Set the file from the directory/name pair
 */

void XGFileSpecifier::SetFile(XGDirectory &dir, char *name)
{
	Assert(dir.fInit);
	
#if OPT_MACOS
	fFileName.vRefNum = dir.fVolRefNum;
	fFileName.parID = dir.fDirID;
	cvtc2p(fFileName.name,name);
#endif

#if OPT_WINOS || OPT_XWIN
	char buffer[512];
	char *x;
	
	strcpy(buffer,dir.fPath.Get());
	for (x = buffer; *x; x++) ;
	x--;

#if OPT_WINOS
	if (*x != '\\') strcat(buffer,"\\");
#else
	if (*x != '/') strcat(buffer,"/");
#endif

	strcat(buffer,name);
	fFileName = buffer;
#endif
}

/*$	XGFileSpecifier::GetName
 *
 *		Get the file name of this file. This extracts just the file
 *	name for naming windows
 */

void XGFileSpecifier::GetName(char *name) const
{
#if OPT_MACOS
	cvtp2c(name,fFileName.name);
#else
	const char *c,*d;

	d = fFileName.Get();
	for (c = d; *c != 0; c++) {
#if OPT_WINOS
		if ((*c == ':') || (*c == '\\')) d = c+1;
#else
		if (*c == '/') d = c+1;
#endif
	}
	
	strcpy(name,d);
#endif
}


void XGFileSpecifier::GetBaseName( char *name, int maxext ) const
{
	this->GetName( name );
	int len = strlen( name ) - 1;
	for (int i = len - 1; i > 0 && i >= len - maxext; --i) {
		if (name[i] == '.') {
			name[i] = 0;
			return;
		}
	}
}


void XGFileSpecifier::GetExtension( char *ext, int maxext ) const
{
	char name[256];
	GetName( name );
	char *ptr = strrchr( name, '.' );
	if (!ptr) {
		*ext = 0;
		return;
	}
	int len = strlen(name);
	if (ptr < name + len - 1 - maxext) {
		*ext = 0;
		return;
	}
	ptr += 1;
	do {
		*ext++ = tolower(*ptr);
	} while (*ptr++);
}


/*$	XGFileSpecifier::GetDirectory
 *
 *		Get the directory that this file is contained as a directory object.
 *	This overrides the directory object provided.
 */

void XGFileSpecifier::GetDirectory(XGDirectory &dir) const
{
#if OPT_MACOS
	dir.fInit = true;
	dir.fVolRefNum = fFileName.vRefNum;
	dir.fDirID = fFileName.parID;
#endif

#if OPT_WINOS || OPT_XWIN
	char *c,*d;
	char buffer[512];

	strcpy(buffer,fFileName.Get());
	d = buffer;
	for (c = d; *c != 0; c++) {
#if OPT_WINOS
		if ((*c == ':') || (*c == '\\')) d = c;
#else
		if (*c == '/') d = c;
#endif
	}
	
#if OPT_WINOS
	if (*d == '\\') {
		if (*(d-1) != ':') d--;
	}
#endif

#if OPT_XWIN
	if (*d == '/') {
		if (d != fFileName.Get()) d--;
	}
#endif

	d[1] = 0;
	dir.fPath = buffer;
	dir.fInit = true;
#endif
}


/************************************************************************/
/*																		*/
/*	File Dialogs (Mac)													*/
/*																		*/
/************************************************************************/

#if OPT_MACOS

#if TARGET_CARBON
pascal void XGFileSpecifier::GEventProc(NavEventCallbackMessage sel, NavCBRecPtr callback, void *)
{
	if (callback->eventData.eventDataParms.event) {
	    WindowPtr window = (WindowPtr)(callback->eventData.eventDataParms.event->message);
	    switch (sel)
	    {
	        case kNavCBEvent:
	            switch (((callback->eventData).eventDataParms).event->what)
	            {
	                case updateEvt:
	                	XGAppCore::GetApplication()->DoUpdate(window);
	                    break;
	            }
	            break;
	    }
	}
}
#endif

/*$	XGFile::OpenDialog
 *
 *		Open the file by running the file open dialog. This returns NULL
 *	if there was an error or the file open dialog was canceled. The
 *	Macintosh version of this takes as it's arguments a 'paramResID',
 *	which refers to an 'Open' resource which specifies the file types
 *	that can be opened.
 *
 *		On Windows, the param string is a string giving the file types
 *	which can be opened. The parameters should be of the format:
 *
 *			(description) '|' *.ext '|' (description) '|' *.ext ...
 *
 *		The second parameter is the default file extension to be used
 *	if one is not provided in the open dialog box.
 *
 *		On Unix, the param string is a comma-separated list of
 *	extensions which can be opened.
 */

XGFileSpecifier *XGFileSpecifier::OpenDialog(short paramResID)
{
#if TARGET_CARBON
	NavReplyRecord nreply;

    XGFileSpecifier* result = 0;
	FSSpec fspec;
	OSErr err;
	long i;
	NavDialogOptions opt;
	long count;
    AEKeyword theKeyword;
    DescType actualType;
    Size actualSize;
    bool retval;
	NavTypeList **nlist;
		
	if (GEventProcPtr == NULL) GEventProcPtr = NewNavEventUPP(GEventProc);
	
	/*
	 *	Set up the options, the file list, and specify that the
	 *	user can only open one file here.
	 */
	
	err = NavGetDefaultDialogOptions(&opt);
	if (err) return false;
	opt.dialogOptionFlags &= ~kNavAllowMultipleFiles;
	opt.dialogOptionFlags |= kNavNoTypePopup;
	
	//if (fLength == -1) {
		opt.dialogOptionFlags |= kNavAllFilesInPopup;
		nlist = NULL;
	//} else {
	//	nlist = (NavTypeList **)NewHandle(sizeof(NavTypeList) + sizeof(OSType) * (fLength - 1));
	//	(*nlist)->componentSignature = kNavGenericSignature;
	//	(*nlist)->reserved = 0;
	//	(*nlist)->osTypeCount = fLength;
	//	for (i = 0; i < fLength; i++) (*nlist)->osType[i] = fList[i];
	//}
	
	/*
	 *	Use the basic navigation services open method
	 */

	err = NavGetFile(NULL,&nreply,&opt,GEventProcPtr,NULL,NULL,nlist,NULL);
	if (nlist) DisposeHandle((Handle)nlist);

	retval = false;
	if (!err && nreply.validRecord) {
		err = AECountItems(&(nreply.selection),&count);
		if (!err) {
			err = AEGetNthPtr(&nreply.selection,
							  1,
							  typeFSS,
							  &theKeyword,
							  &actualType,
							  &fspec,
							  sizeof(fspec),
							  &actualSize);
			if (!err) {
				retval = true;
				result = new XGFileSpecifier(fspec);
			}
		}
	}
	NavDisposeReply(&nreply);
	return result;
#else
	SFTypeList list;
	Handle h;
	short i,len;
	FSSpec f;
	OSErr err;
	extern bool _GHasFSSpec;
	
	/*
	 *	Get resource information
	 */
	
	h = Get1Resource('Open',paramResID);
	Assert(h);
	len = *(short *)*h;
	if (len > 4) len = 4;
	for (i = 0; i < len; i++) {
		list[i] = ((OSType *)(2 + *h))[i];
	}
	if (len == 0) len = -1;
	
	/*
	 *	Get the specified file from the user
	 */
	
	if (_GHasFSSpec) {
		StandardFileReply sfReply;

		/*
		 *	Sys7 version of this uses StandardGetFile() routine
		 *	to get file specifier
		 */
		
		StandardGetFile(NULL,len,list,&sfReply);
		if (!sfReply.sfGood) return NULL;
		f = sfReply.sfFile;
	} else {
		Point pt;
		SFReply reply;
		
		/*
		 *	Sys6 (and earlier) version uses SFGetFile and
		 *	converts the wdRefNum to a FileSpec
		 */
		
		pt.h = 75, pt.v = 75;
		SFGetFile(pt,NULL,NULL,len,list,NULL,&reply);
		if (!reply.good) return NULL;
		
		pstrcpy(f.name,reply.fName);
		err = GetWDInfo(reply.vRefNum,&f.vRefNum,&f.parID,NULL);
		Assert(err == noErr);
	}
	
	/*
	 *	Create the file and open it
	 */
	
	return new XGFileSpecifier(f);
#endif
/*	file = new XGFile;
	file->Open(&spec,rwflag);
	return file; */
}

/*$	XGFile::SaveDialog
 *
 *		Run a save dialog and create a new file with the specified name.
 *
 *		On the Macintosh, this specifies a resource ID of a 'Save' resource
 *	with the default prompt, the creator and file type specifiers. The creator
 *	and file type specifier is used when the file is created.
 *
 *		On Windows, this specifies the file types, in the same format
 *	string used by OpenDialog. On Unix, this specifies the default extension
 *	to be used, as well as a prompt string.
 */
 
XGFileSpecifier *XGFileSpecifier::SaveDialog(short paramResID, OSType c, OSType t, char *defname)
{
#pragma unused(c)
#pragma unused(t)

#if TARGET_CARBON
    XGFileSpecifier* result = 0;

	pchar pbuffer[256];
	pchar dbuffer[256];
	NavReplyRecord nreply;

	FSSpec fspec;
	OSErr err;
	NavDialogOptions opt;
	long count;
    AEKeyword theKeyword;
    DescType actualType;
    Size actualSize;
    bool retval;
	
	char prompt[255];
	char defaultname[255];
	strcpy( prompt, "" );
	strcpy( defaultname, "" );
	cvtc2p(pbuffer,prompt);
	cvtc2p(dbuffer,defaultname);

	/*
	 *	Use the basic navigation services save method
	 */

	if (GEventProcPtr == NULL) GEventProcPtr = NewNavEventUPP(GEventProc);
	
	/*
	 *	Set up the options, the file list, and specify that the
	 *	user can only open one file here.
	 */
	
	err = NavGetDefaultDialogOptions(&opt);
	if (err) return false;
	opt.dialogOptionFlags &= ~kNavAllowMultipleFiles;
	opt.dialogOptionFlags |= kNavNoTypePopup;
	opt.dialogOptionFlags |= kNavDontAutoTranslate;
	cvtc2p(opt.savedFileName,defaultname);
//		cvtc2p(opt.message,prompt);
	
	/*
	 *	Use the basic navigation services save method
	 */

	err = NavPutFile(NULL,&nreply,&opt,GEventProcPtr,kNavGenericSignature,kNavGenericSignature,NULL);

	retval = false;
	if (!err && nreply.validRecord) {
		err = AECountItems(&(nreply.selection),&count);
		if (!err) {
			err = AEGetNthPtr(&nreply.selection,
							  1,
							  typeFSS,
							  &theKeyword,
							  &actualType,
							  &fspec,
							  sizeof(fspec),
							  &actualSize);
			if (!err) {
				retval = true;
				result = new XGFileSpecifier(fspec);
			}
		}
	}
	NavDisposeReply(&nreply);
	return result;
#else
	unsigned char prompt[256];
	unsigned char str[64];
	Handle h;
	FSSpec f;
	OSErr err;
	extern bool _GHasFSSpec;
	
	h = Get1Resource('Save',paramResID);
	Assert(h);
	pstrcpy(prompt,(unsigned char *)*h);
	ReleaseResource(h);

	/*
	 *	Get the specified file from the user
	 */
	
	cvtc2p(str,defname);
	if (_GHasFSSpec) {
		StandardFileReply sfReply;

		/*
		 *	Sys7 version of this uses StandardPutFile routine
		 *	to get file specifier
		 */
		
		StandardPutFile(prompt,str,&sfReply);
		if (!sfReply.sfGood) return NULL;
		f = sfReply.sfFile;
	} else {
		Point pt;
		SFReply reply;
		
		/*
		 *	Sys6 (and earlier) gets the file name and
		 *	converts into a file specifier
		 */
		
		pt.h = 75, pt.v = 75;
		SFPutFile(pt,prompt,str,NULL,&reply);
		if (!reply.good) return NULL;
		
		pstrcpy(f.name,reply.fName);
		err = GetWDInfo(reply.vRefNum,&f.vRefNum,&f.parID,NULL);
		Assert(err == noErr);
	}
	
	/*
	 *	Create the file and open it
	 */
	
	return new XGFileSpecifier( f );
/*	spec = f;
	file = new XGFile;
	file->Create(&spec,c,t);
	return file; */
#endif
}

#endif



/************************************************************************/
/*																		*/
/*	File Dialogs (Win)													*/
/*																		*/
/************************************************************************/

#if OPT_WINOS

/*	XGFileSpecifier::OpenDialog
 *
 *		Open the file by running the file open dialog. This returns NULL
 *	if there was an error or the file open dialog was canceled. The
 *	Macintosh version of this takes as it's arguments a 'paramResID',
 *	which refers to an 'Open' resource which specifies the file types
 *	that can be opened.
 *
 *		On Windows, the param string is a string giving the file types
 *	which can be opened. The parameters should be of the format:
 *
 *			(description) '|' *.ext '|' (description) '|' *.ext ...
 *
 *		The second parameter is the default file extension to be used
 *	if one is not provided in the open dialog box.
 *
 *		On Unix, the param string is a comma-separated list of
 *	extensions which can be opened.
 */

XGFileSpecifier *XGFileSpecifier::OpenDialog(char *type, char *ext)
{
	OPENFILENAME ofn;
	char buffer[256];
	char name[256];
	char *c;
	
	memset(&ofn,0,sizeof(ofn));
	ofn.lStructSize = sizeof(ofn);
	ofn.hwndOwner = _GMainWindow;
	ofn.hInstance = _GInstance;
	ofn.lpstrDefExt = ext;
	
	memset(buffer, 0, sizeof(buffer));
	strcpy(buffer,type);
	for (c = buffer; *c; c++) if (*c == '|') *c = '\0';
	ofn.lpstrFilter = buffer;
	
	name[0] = 0;
	ofn.lpstrFile = name;
	ofn.nMaxFile = 256;
	ofn.Flags = OFN_FILEMUSTEXIST | OFN_LONGNAMES;

	if (GetOpenFileName(&ofn)) {
		return new XGFileSpecifier( name );
	/*	XGFileSpecifier fspec;
		f = new XGFile;
		fspec = name;
		f->Open(&fspec,rwflag);
		return f; */
	} else return NULL;
}

/*	XGFile::SaveDialog
 *
 *		Run a save dialog and create a new file with the specified name.
 *
 *		On the Macintosh, this specifies a resource ID of a 'Save' resource
 *	with the default prompt, the creator and file type specifiers. The creator
 *	and file type specifier is used when the file is created.
 *
 *		On Windows, this specifies the file types, in the same format
 *	string used by OpenDialog. On Unix, this specifies the default extension
 *	to be used, as well as a prompt string.
 */

XGFileSpecifier *XGFileSpecifier::SaveDialog(char *type, char *ext, char *defname)
{
	OPENFILENAME ofn;
	char buffer[256];
	char name[256];
	char *c;
	
	memset(&ofn,0,sizeof(ofn));
	ofn.lStructSize = sizeof(ofn);
	ofn.hwndOwner = _GMainWindow;
	ofn.hInstance = _GInstance;
	ofn.lpstrDefExt = ext;
	
	memset(buffer, 0, sizeof(buffer));
	strcpy(buffer,type);
	for (c = buffer; *c; c++) if (*c == '|') *c = '\0';
	ofn.lpstrFilter = buffer;
	
	strcpy(name,defname);
	ofn.lpstrFile = name;
	ofn.nMaxFile = 256;
	ofn.Flags = OFN_OVERWRITEPROMPT | OFN_LONGNAMES;

	if (GetSaveFileName(&ofn)) {
		return new XGFileSpecifier( name );
/*		XGFileSpecifier fspec;
		
		f = new XGFile;
		fspec = name;
		f->Create(&fspec);
		return f; */
	} else return NULL;
}

#endif


bool XGFileSpecifier::ExchangeFileContents( XGFileSpecifier &file1, XGFileSpecifier &file2 )
{
#if OPT_MACOS
	return (FSpExchangeFiles( &file1.fFileName, &file2.fFileName ) == noErr);
#else
	XGFileSpecifier temp;
	temp = file1.TempInSameDirectory();
	temp.DeleteFile();
	if (RenameFile( file1, temp )) {
		if (RenameFile( file2, file1)) {
			if (RenameFile( temp, file2 )) {
				// success
				return true;
			} else {
				RenameFile( file1, file2 );
				RenameFile( temp, file1 );
				return false;
			}
		} else {
			RenameFile( temp, file1 );
			return false;
		}
	} else return false;
#endif
}

bool XGFileSpecifier::RenameFile(const XGFileSpecifier &file1, const XGFileSpecifier &file2 )
{
#pragma unused(file1)
#pragma unused(file2)

#if OPT_WINOS
	return MoveFile(file1.fFileName.Get(), file2.fFileName.Get());
#else
	return false;
#endif
}

#if OPT_MACOS
XGFileSpecifier XGFileSpecifier::TempInSameDirectory()
{
	int counter = 0;
	char name[256];
	FSSpec fs;
	for (;;) {
		sprintf( name+1, "temp-%x", ++counter );
		name[0] = strlen(name+1);
		OSStatus err;
		
		err = FSMakeFSSpec( fFileName.vRefNum, fFileName.parID, (unsigned char *)name, &fs );
		if (err == 0) continue;
		if (err != fnfErr) {
			fs.vRefNum = 0;
			fs.parID = 0;
			fs.name[0] = 0;
			break;
		}
		
		err = FSpCreate( &fs, 'temp', 'temp', smSystemScript );
		if (err == 0) break;
		if (err == dupFNErr) continue;
		FSpDelete( &fs );
		fs.vRefNum = 0;
		fs.parID = 0;
		fs.name[0] = 0;
		break;
	}
	
	return XGFileSpecifier( fs );
}
#else
XGFileSpecifier XGFileSpecifier::TempInSameDirectory()
{
	XGDirectory dir;
	this->GetDirectory( dir );
	
	int counter=0;
	char name[256];
	
	for(;;)
	{
		sprintf(name, "temp-%x", ++counter);
		XGFileSpecifier temp;
		temp.SetFile( dir, name );
		FILE *fp = temp.OpenFile( "rb" );
		if (!fp) {
			fp = temp.OpenFile( "wb" );
			if (fp) {
				fclose( fp );
				return temp;
			}
		}
		fclose(fp);
	}
}
#endif

#if OPT_MACOS
void XGFileSpecifier::DeleteFile()
{
	FSpDelete( &this->fFileName );
}
#elif OPT_WINOS
void XGFileSpecifier::DeleteFile()
{
	::DeleteFile( fFileName.Get() );	
}
#else
/* -- ALAN: added missing OPT_XWIN function -- */
void XGFileSpecifier::DeleteFile()
{
	remove( fFileName.Get() );	
}
#endif




#if OPT_MACOS
void XGFileSpecifier::SetTypeCreator( OSType type, OSType creator )
{
	FInfo fi;
	if (FSpGetFInfo( &fFileName, &fi ) == noErr) {
		fi.fdType = type;
		fi.fdCreator = creator;
		FSpSetFInfo( &fFileName, &fi );
	}
}


OSType XGFileSpecifier::GetType()
{
	FInfo fi;
	if (FSpGetFInfo( &fFileName, &fi ) == noErr) return fi.fdType;
	return 0;
}


OSType XGFileSpecifier::GetCreator()
{
	FInfo fi;
	if (FSpGetFInfo( &fFileName, &fi ) == noErr) return fi.fdCreator;
	return 0;
}
#endif






/* -- ALAN: pulled this stuff in from elsewhere, as this functionality was missing -- */ 
/************************************************************************/
/*																		*/
/*	File Dialogs (XWin)													*/
/*																		*/
/************************************************************************/

#if OPT_XWIN

/*
 *	Globals
 */

static char			*GExtension;
static char			*GFilter;
static bool			GSaveFlag;			// save file (new file name) dialog?
static XGStringList	*GList;

/*
 *	Built-in dialogs
 */

static unsigned char GOpenDialogData[] = {
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x60, /* 00000000: .......` */
    0x00, 0x00, 0x00, 0x50, 0x64, 0x6C, 0x6F, 0x67, /* 00000008: ...Pdlog */
    0x00, 0x00, 0x00, 0x12, 0x00, 0x00, 0x00, 0x00, /* 00000010: ........ */
    0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000018: ........ */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000020: ........ */
    0x00, 0x00, 0x00, 0x00, 0x01, 0x48, 0x00, 0x00, /* 00000028: .....H.. */
    0x00, 0xEC, 0x00, 0x00, 0x01, 0x48, 0x00, 0x00, /* 00000030: .....H.. */
    0x00, 0xEC, 0x00, 0x00, 0x01, 0x48, 0x00, 0x00, /* 00000038: .....H.. */
    0x00, 0xEC, 0x00, 0x00, 0x01, 0x48, 0x00, 0x00, /* 00000040: .....H.. */
    0x00, 0xEC, 0x09, 0x4F, 0x70, 0x65, 0x6E, 0x20, /* 00000048: ...Open. */
    0x46, 0x69, 0x6C, 0x65, 0x00, 0x00, 0x00, 0x00, /* 00000050: File.... */
    0x00, 0x00, 0x00, 0x64, 0x00, 0x00, 0x00, 0x65, /* 00000058: ...d...e */
    0x00, 0x00, 0x00, 0x3A, 0x00, 0x00, 0x00, 0x00, /* 00000060: ...:.... */
    0x00, 0x00, 0x00, 0x2A, 0x70, 0x6F, 0x70, 0x6D, /* 00000068: ...*popm */
    0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, /* 00000070: ........ */
    0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, /* 00000078: ........ */
    0x00, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00, 0x0C, /* 00000080: ........ */
    0x00, 0x00, 0x00, 0xC8, 0x00, 0x00, 0x00, 0x24, /* 00000088: .......$ */
    0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, /* 00000090: ........ */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x44, 0x00, 0x00, /* 00000098: .....D.. */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x34, 0x6C, 0x69, /* 000000A0: .....4li */
    0x73, 0x74, 0x00, 0x00, 0x00, 0x02, 0x00, 0x00, /* 000000A8: st...... */
    0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, /* 000000B0: ........ */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x00, 0x00, /* 000000B8: ........ */
    0x00, 0x28, 0x00, 0x00, 0x00, 0xEC, 0x00, 0x00, /* 000000C0: .(...... */
    0x00, 0xC8, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, /* 000000C8: ........ */
    0x00, 0x00, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, /* 000000D0: ........ */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 000000D8: ........ */
    0x00, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 000000E0: .@...... */
    0x00, 0x30, 0x74, 0x65, 0x78, 0x74, 0x00, 0x00, /* 000000E8: .0text.. */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, /* 000000F0: ........ */
    0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 000000F8: ........ */
    0x00, 0x10, 0x00, 0x00, 0x00, 0xD0, 0x00, 0x00, /* 00000100: ........ */
    0x00, 0x40, 0x00, 0x00, 0x00, 0xE0, 0xFF, 0xFF, /* 00000108: .@...... */
    0xFF, 0xFF, 0x00, 0x00, 0x05, 0x26, 0x4F, 0x70, /* 00000110: .....&Op */
    0x65, 0x6E, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000118: en...... */
    0x00, 0x52, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000120: .R...... */
    0x00, 0x42, 0x65, 0x64, 0x69, 0x74, 0x00, 0x00, /* 00000128: .Bedit.. */
    0x00, 0x03, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, /* 00000130: ........ */
    0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000138: ........ */
    0x00, 0x48, 0x00, 0x00, 0x00, 0xCD, 0x00, 0x00, /* 00000140: .H...... */
    0x00, 0xEC, 0x00, 0x00, 0x00, 0xE3, 0xFF, 0xFF, /* 00000148: ........ */
    0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, /* 00000150: ........ */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000158: ........ */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000160: ........ */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000168: ........ */
    0x00, 0x00, 0x00, 0x44, 0x00, 0x00, 0x00, 0x00, /* 00000170: ...D.... */
    0x00, 0x00, 0x00, 0x34, 0x70, 0x62, 0x74, 0x6E, /* 00000178: ...4pbtn */
    0x00, 0x00, 0x00, 0x64, 0x00, 0x00, 0x00, 0x00, /* 00000180: ...d.... */
    0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, /* 00000188: ........ */
    0x00, 0x00, 0x00, 0xFC, 0x00, 0x00, 0x00, 0x0C, /* 00000190: ........ */
    0x00, 0x00, 0x01, 0x3E, 0x00, 0x00, 0x00, 0x28, /* 00000198: ...>...( */
    0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, /* 000001A0: ........ */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x4F, /* 000001A8: .......O */
    0x70, 0x65, 0x6E, 0x00, 0x00, 0x00, 0x00, 0x46, /* 000001B0: pen....F */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x36, /* 000001B8: .......6 */
    0x70, 0x62, 0x74, 0x6E, 0x00, 0x00, 0x00, 0x65, /* 000001C0: pbtn...e */
    0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, /* 000001C8: ........ */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFC, /* 000001D0: ........ */
    0x00, 0x00, 0x00, 0x2C, 0x00, 0x00, 0x01, 0x3E, /* 000001D8: ...,...> */
    0x00, 0x00, 0x00, 0x48, 0xFF, 0xFF, 0xFF, 0xFF, /* 000001E0: ...H.... */
    0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 000001E8: ........ */
    0x00, 0x00, 0x06, 0x43, 0x61, 0x6E, 0x63, 0x65, /* 000001F0: ...Cance */
    0x6C, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 000001F8: l....... */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x34, 0x70, 0x62, /* 00000200: .....4pb */
    0x74, 0x6E, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, /* 00000208: tn...... */
    0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, /* 00000210: ........ */
    0x00, 0x00, 0x00, 0x00, 0x00, 0xFC, 0x00, 0x00, /* 00000218: ........ */
    0x00, 0x58, 0x00, 0x00, 0x01, 0x3C, 0x00, 0x00, /* 00000220: .X...<.. */
    0x00, 0x74, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, /* 00000228: .t...... */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000230: ........ */
    0x04, 0x48, 0x6F, 0x6D, 0x65, 0x00              /* 00000238: .Home.   */
}; /* _R56526573_0085 */

static unsigned char GSaveDialogData[] = {
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x60, /* 00000000: .......` */
    0x00, 0x00, 0x00, 0x50, 0x64, 0x6C, 0x6F, 0x67, /* 00000008: ...Pdlog */
    0x00, 0x00, 0x00, 0x12, 0x00, 0x00, 0x00, 0x00, /* 00000010: ........ */
    0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000018: ........ */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000020: ........ */
    0x00, 0x00, 0x00, 0x00, 0x01, 0x48, 0x00, 0x00, /* 00000028: .....H.. */
    0x00, 0xEC, 0x00, 0x00, 0x01, 0x48, 0x00, 0x00, /* 00000030: .....H.. */
    0x00, 0xEC, 0x00, 0x00, 0x01, 0x48, 0x00, 0x00, /* 00000038: .....H.. */
    0x00, 0xEC, 0x00, 0x00, 0x01, 0x48, 0x00, 0x00, /* 00000040: .....H.. */
    0x00, 0xEC, 0x09, 0x53, 0x61, 0x76, 0x65, 0x20, /* 00000048: ...Save. */
    0x46, 0x69, 0x6C, 0x65, 0x00, 0x00, 0x00, 0x00, /* 00000050: File.... */
    0x00, 0x00, 0x00, 0x64, 0x00, 0x00, 0x00, 0x65, /* 00000058: ...d...e */
    0x00, 0x00, 0x00, 0x3A, 0x00, 0x00, 0x00, 0x00, /* 00000060: ...:.... */
    0x00, 0x00, 0x00, 0x2A, 0x70, 0x6F, 0x70, 0x6D, /* 00000068: ...*popm */
    0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, /* 00000070: ........ */
    0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, /* 00000078: ........ */
    0x00, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00, 0x0C, /* 00000080: ........ */
    0x00, 0x00, 0x00, 0xC8, 0x00, 0x00, 0x00, 0x24, /* 00000088: .......$ */
    0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, /* 00000090: ........ */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x44, 0x00, 0x00, /* 00000098: .....D.. */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x34, 0x6C, 0x69, /* 000000A0: .....4li */
    0x73, 0x74, 0x00, 0x00, 0x00, 0x02, 0x00, 0x00, /* 000000A8: st...... */
    0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, /* 000000B0: ........ */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x00, 0x00, /* 000000B8: ........ */
    0x00, 0x28, 0x00, 0x00, 0x00, 0xEC, 0x00, 0x00, /* 000000C0: .(...... */
    0x00, 0xC8, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, /* 000000C8: ........ */
    0x00, 0x00, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, /* 000000D0: ........ */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 000000D8: ........ */
    0x00, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 000000E0: .@...... */
    0x00, 0x30, 0x74, 0x65, 0x78, 0x74, 0x00, 0x00, /* 000000E8: .0text.. */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, /* 000000F0: ........ */
    0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 000000F8: ........ */
    0x00, 0x10, 0x00, 0x00, 0x00, 0xD0, 0x00, 0x00, /* 00000100: ........ */
    0x00, 0x40, 0x00, 0x00, 0x00, 0xE0, 0xFF, 0xFF, /* 00000108: .@...... */
    0xFF, 0xFF, 0x00, 0x00, 0x05, 0x26, 0x53, 0x61, /* 00000110: .....&Sa */
    0x76, 0x65, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000118: ve...... */
    0x00, 0x52, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000120: .R...... */
    0x00, 0x42, 0x65, 0x64, 0x69, 0x74, 0x00, 0x00, /* 00000128: .Bedit.. */
    0x00, 0x03, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, /* 00000130: ........ */
    0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000138: ........ */
    0x00, 0x48, 0x00, 0x00, 0x00, 0xCD, 0x00, 0x00, /* 00000140: .H...... */
    0x00, 0xEC, 0x00, 0x00, 0x00, 0xE3, 0xFF, 0xFF, /* 00000148: ........ */
    0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, /* 00000150: ........ */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000158: ........ */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000160: ........ */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000168: ........ */
    0x00, 0x00, 0x00, 0x44, 0x00, 0x00, 0x00, 0x00, /* 00000170: ...D.... */
    0x00, 0x00, 0x00, 0x34, 0x70, 0x62, 0x74, 0x6E, /* 00000178: ...4pbtn */
    0x00, 0x00, 0x00, 0x64, 0x00, 0x00, 0x00, 0x00, /* 00000180: ...d.... */
    0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, /* 00000188: ........ */
    0x00, 0x00, 0x00, 0xFC, 0x00, 0x00, 0x00, 0x0C, /* 00000190: ........ */
    0x00, 0x00, 0x01, 0x3E, 0x00, 0x00, 0x00, 0x28, /* 00000198: ...>...( */
    0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, /* 000001A0: ........ */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x53, /* 000001A8: .......S */
    0x61, 0x76, 0x65, 0x00, 0x00, 0x00, 0x00, 0x46, /* 000001B0: ave....F */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x36, /* 000001B8: .......6 */
    0x70, 0x62, 0x74, 0x6E, 0x00, 0x00, 0x00, 0x65, /* 000001C0: pbtn...e */
    0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, /* 000001C8: ........ */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFC, /* 000001D0: ........ */
    0x00, 0x00, 0x00, 0x2C, 0x00, 0x00, 0x01, 0x3E, /* 000001D8: ...,...> */
    0x00, 0x00, 0x00, 0x48, 0xFF, 0xFF, 0xFF, 0xFF, /* 000001E0: ...H.... */
    0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 000001E8: ........ */
    0x00, 0x00, 0x06, 0x43, 0x61, 0x6E, 0x63, 0x65, /* 000001F0: ...Cance */
    0x6C, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 000001F8: l....... */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x34, 0x70, 0x62, /* 00000200: .....4pb */
    0x74, 0x6E, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, /* 00000208: tn...... */
    0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, /* 00000210: ........ */
    0x00, 0x00, 0x00, 0x00, 0x00, 0xFC, 0x00, 0x00, /* 00000218: ........ */
    0x00, 0x58, 0x00, 0x00, 0x01, 0x3C, 0x00, 0x00, /* 00000220: .X...<.. */
    0x00, 0x74, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, /* 00000228: .t...... */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000230: ........ */
    0x04, 0x48, 0x6F, 0x6D, 0x65, 0x00              /* 00000238: .Home.   */
}; /* _R56526573_0086 */

/*	SortList
 *
 *		Do a single sort for the qsort object. 
 */

static int SortList(const void *a, const void *b)
{
	short aa = *(short *)a;
	short bb = *(short *)b;
	char buf1[256];
	char buf2[256];
	int i;
	
	GList->GetString(aa,buf1);
	GList->GetString(bb,buf2);
	i = strcmp(buf1,buf2);
	if (i < 0) return -1;
	if (i > 0) return 1;
	if (aa < bb) return -1;
	if (aa > bb) return 1;
	return 0;
}

/*	LoadCWD
 *
 *		Load the current working directory into the popup menu control
 */
#include <XPopup.h>
#include <unistd.h>
#include <XList.h>
static void LoadCWD(XGPopControl *pop)
{
	char buffer[512];
	char *a,*b;
	int count;
	
	/*
	 *	Get the current working directory
	 */
	
	getcwd(buffer,512);
	
	/*
	 *	Clear the popup menu
	 */
	
	count = 0;
	pop->DeleteAll();
	pop->Insert(0,"/");					// root directory
	a = buffer+1;
	while (*a) {
		b = buffer;
		while ((*a != '/') && (*a != 0)) *b++ = *a++;
		if (*a) a++;
		*b = 0;
		pop->Insert(++count,buffer);
		if (*a == 0) break;
	}
	pop->SetValue(count);
}

/*	FilterName
 *
 *		Filter the file name. Return true if this should be skipped.
 */

static bool FilterName(char *name)
{
	char *x,*y,*a,*b;
	
	if (GFilter == NULL) return false;

	y = NULL;
	for (x = name; *x; x++) {
		if (*x == '.') y = x+1;
	}
	if (!y) y = x;
	
	x = GFilter;
	while (*x) {
		for (a = y, b = x; *a; a++, b++) {
			if (*a != *b) break;
		}
		if ((*a == 0) && ((*b == ',') || (*b == 0))) return false;
		
		while ((*x) && (*x != ',')) x++;
		if (*x == 0) break;
		x++;
	}
	return true;
}

/*	LoadDir
 *
 *		Load the contents of the directory into the drop list
 */
#include <sys/stat.h>
static bool LoadDir(XGListControl *list)
{
	XGStringList slist;
	char buffer[512];
	DIR *dir;
	struct dirent *file;
	struct stat filestat;
	short *lsort;
	int i;
	int len;
	
	GList = &slist;
	
	/*
	 *	Open directory and load list
	 */
	
	getcwd(buffer,512);
	dir = opendir(buffer);
	if (dir == NULL) return false;
	
	while (NULL != (file = readdir(dir))) {
		strcpy(buffer,file->d_name);
		if (!strcmp(buffer,".") || !strcmp(buffer,"..")) continue;
		stat(buffer,&filestat);
		if (S_ISDIR(filestat.st_mode)) {
			strcat(buffer,"/");
		} else if (FilterName(buffer)) continue;
		
		slist.Append(buffer);
	}
	closedir(dir);
	
	/*
	 *	Empty list?
	 */
	
	len = (int)slist.GetNumStrings();
	if (len == 0) {
		list->DeleteAll();
		return true;
	}
	
	/*
	 *	Sort the list
	 */
	
	lsort = (short *)malloc(sizeof(short) * len);
	for (i = 0; i < len; i++) lsort[i] = i;
	qsort((void *)lsort,len,sizeof(short),SortList);
	
	/*
	 *	Load the list
	 */
	
	list->SetUpdate(false);
	list->DeleteAll();
	for (i = 0; i < len; i++) {
		slist.GetString(lsort[i],buffer);
		list->Insert(i,buffer);
	}
	list->SetUpdate(true);
	
	free((void *)lsort);
}

/*	SetExtension
 *
 *		Set the file extension if it doesn't match the default
 *	extension
 */

static void SetExtension(char *c)
{
	char *x,*y;
	
	if (GExtension == NULL) return;
	
	y = NULL;
	for (x = c; *x; x++) {
		if (*x == '.') y = x+1;
	}
	if (y) {
		if (!strcmp(y,GExtension)) return;
	}
	strcat(c,".");
	strcat(c,GExtension);
}

/*	DialogProc
 *
 *		This is the proc to handle the open and save dialogs.
 */
#include <XDialog.h>
#include <XTextControls.h>
#include <XMessages.h>
static long DialogProc(XGDialog *dlog, void *data, long msg, long arg, void *parg)
{
	XGListControl *c;
	XGEditText *e;
	char buffer[512];
	long sel;
	char *end;
	struct stat status;
	XGPopControl *ctrl;
	char str[256];
	long i,len;
	
	switch (msg) {
		case KXGDialogMessage:
			if (arg == KXGDMInitDialog) {
				/*
				 *	Load the contents of this thing
				 */
				
				LoadCWD(dynamic_cast<XGPopControl *>(dlog->FindViewByID(1)));
				LoadDir(dynamic_cast<XGListControl *>(dlog->FindViewByID(2)));

				e = dynamic_cast<XGEditText *>(dlog->FindViewByID(3));
				if (GSaveFlag) e->SetTitle((char *)data);
				else e->SetTitle("");
				break;
			}
			return -1;
			
		case KXGListSelect:
			/*
			 *	First, emulate the list value select
			 */
			
			if (-1 == DialogProc(dlog,data,KXGListValue,2,NULL)) return -1;
			dlog->GetText(3,sizeof(buffer),buffer);
			if (buffer[0]) {
				/*
				 *	Second, emulate the 'ok' click
				 */
				
				return DialogProc(dlog,data,KXGControlClick,100,NULL);
			}
			break;
						
		case KXGListValue:
			c = dynamic_cast<XGListControl *>(dlog->FindViewByID(2));
			if (c) {
				sel = c->GetValue();
				if (sel == -1) {
					if (GSaveFlag) strcpy(buffer,(char *)data);
					else buffer[0] = 0;
				}
				else c->GetItem(sel,buffer);
				e = dynamic_cast<XGEditText *>(dlog->FindViewByID(3));
				for (end = buffer; *end; end++) {
					if (*end == '/') {
						*end = 0;
						break;
					}
				}
				e->SetTitle(buffer);
			}
			return 0;
			
		case KXGPopValue:
			ctrl = dynamic_cast<XGPopControl *>(dlog->FindViewByID(1));
			if (ctrl) {
				/*
				 *	Figure out how far up to go
				 */
				
				len = ctrl->GetValue();
				if (len == ctrl->Length()-1) return 0;
				strcpy(buffer,"/");
				for (i = 1; i < len; i++) {
					ctrl->GetItem(i,str);
					strcat(buffer,str);
					strcat(buffer,"/");
				}
				ctrl->GetItem(len,str);
				strcat(buffer,str);
				
				/*
				 *	Now set the directory
				 */
				
				if (!chdir(buffer)) {
					LoadCWD(dynamic_cast<XGPopControl *>(dlog->FindViewByID(1)));
					LoadDir(dynamic_cast<XGListControl *>(dlog->FindViewByID(2)));
				}
			}
			return 0;
	
		case KXGControlClick:
			if (arg == 4) {
				/*
				 *	Go to the user's home directory
				 */
				
				if (!chdir(getenv("HOME"))) {
					e = dynamic_cast<XGEditText *>(dlog->FindViewByID(3));
					LoadCWD(dynamic_cast<XGPopControl *>(dlog->FindViewByID(1)));
					LoadDir(dynamic_cast<XGListControl *>(dlog->FindViewByID(2)));
					if (GSaveFlag) e->SetTitle((char *)data);
					else e->SetTitle("");
				}
				return 0;
			}
			if (arg == 101) {
				dlog->EndDialog(0);
				return 0;
			}
			if (arg == 100) {
				/*
				 *	Is this a directory?
				 */
				
				e = dynamic_cast<XGEditText *>(dlog->FindViewByID(3));
				e->GetTitle(buffer);
				
				if (GSaveFlag) {
					/*
					 *	Add the extension to the save only if the file
					 *	does not exist, or if it's not a directory. If it's
					 *	a directory, we're changing directories. If it's
					 *	a file that exists, we're replacing it...
					 */
					
					if (-1 == stat(buffer,&status)) {
						SetExtension(buffer);
					}
					if (!S_ISDIR(status.st_mode)) {
						SetExtension(buffer);
					}
				}
				
				if (-1 == stat(buffer,&status)) {

					/*
					 *	Does not exist.
					 */
					
					if (!GSaveFlag) {
						Error(KXGFileNoExistError,buffer);
						return 0;
					} else {
						strcpy((char *)data,buffer);
						dlog->EndDialog(1);
						return 0;
					}
				} else {
					/*
					 *	Exists
					 */
					
					if (S_ISDIR(status.st_mode)) {
						/*
						 *	Directory. Change dir down
						 */
						
						if (!chdir(buffer)) {
							LoadCWD(dynamic_cast<XGPopControl *>(dlog->FindViewByID(1)));
							LoadDir(dynamic_cast<XGListControl *>(dlog->FindViewByID(2)));
							if (GSaveFlag) e->SetTitle((char *)data);
							else e->SetTitle("");
						}
						return 0;
					}
					
					/*
					 *	File.
					 */
					
					if (GSaveFlag) {
						if (Caution(KXGFileExistError,buffer)) return 0;
					}
					strcpy((char *)data,buffer);
					dlog->EndDialog(1);
					return 0;
				}
			}
				
			return -1;
			
	}
	return -1;
}

/*	XGFile::OpenDialog
 *
 *		Open the file by running the file open dialog. This returns NULL
 *	if there was an error or the file open dialog was canceled. The
 *	Macintosh version of this takes as it's arguments a 'paramResID',
 *	which refers to an 'Open' resource which specifies the file types
 *	that can be opened.
 *
 *		On Windows, the param string is a string giving the file types
 *	which can be opened. The parameters should be of the format:
 *
 *			(description) '|' *.ext '|' (description) '|' *.ext ...
 *
 *		The second parameter is the default file extension to be used
 *	if one is not provided in the open dialog box.
 *
 *		On Unix, the param string is a comma-separated list of
 *	extensions which can be opened.
 */

 
XGFileSpecifier *XGFileSpecifier::OpenDialog(char *tlist)
{
	char fname[256];

	GSaveFlag = false;
	fname[0] = 0;
	GFilter = tlist;
	if (!XGDialog::ModalDialog(GOpenDialogData,DialogProc,(void *)fname)) {
		return NULL;
	}

	return new XGFileSpecifier( fname );
}

/*	XGFile::SaveDialog
 *
 *		Run a save dialog and create a new file with the specified name.
 *
 *		On the Macintosh, this specifies a resource ID of a 'Save' resource
 *	with the default prompt, the creator and file type specifiers. The creator
 *	and file type specifier is used when the file is created.
 *
 *		On Windows, this specifies the file types, in the same format
 *	string used by OpenDialog. On Unix, this specifies the default extension
 *	to be used, as well as a prompt string.
 */


XGFileSpecifier	*XGFileSpecifier::SaveDialog(char *ext, char *name)
{
	char fname[256];
	
	GSaveFlag = true;
	strcpy(fname,name);
	GFilter = NULL;
	GExtension = ext;
	if (!XGDialog::ModalDialog(GSaveDialogData,DialogProc,(void *)fname)) {
		return NULL;
	}

	return new XGFileSpecifier( fname );
}

#endif

// ALAN: end XWin open/save dialogs